/*
 * base64.c	Functions for base64 (MIME) encoding/decoding
 *
 * Version:	$Id$
 *
 *   This library is free software; you can redistribute it and/or
 *   modify it under the terms of the GNU Lesser General Public
 *   License as published by the Free Software Foundation; either
 *   version 2.1 of the License, or (at your option) any later version.
 *
 *   This library is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
 *   Lesser General Public License for more details.
 *
 *   You should have received a copy of the GNU Lesser General Public
 *   License along with this library; if not, write to the Free Software
 *   Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301, USA
 *
 * Copyright 2007,2012 ARNES <aaa-podpora@arnes.si>, http://www.arnes.si
 */

static const char rcsid[] = "$Id$";

#include <freeradius-devel/autoconf.h>
#include <string.h>

#ifdef HAVE_PTHREAD_H
#include <pthread.h>
pthread_mutex_t fr_base64_mutx = PTHREAD_MUTEX_INITIALIZER;
#endif /* HAVE_PTHREAD_H */

#include <freeradius-devel/libradius.h>
#include <freeradius-devel/base64.h>

static int fr_base64_init = 0;
static const unsigned char fr_base64_enc[] = "ABCDEFGHIJKLMNOPQRSTUVWXYZabcdefghijklmnopqrstuvwxyz0123456789+/";
static unsigned char fr_base64_dec[256];


/* ==========================================================
 Input: binary -> string of binary data (\0x00 also allowed)
        len -> length of data
 Output: Malloced base64 encoded string (printable characters only!)
         Don't forget to free() it after use.
 
   ----------------------------------------------------------
 encoded must be 1/3 longer in size from binary 3 bin -> 4 enc
 11111111 22222222 33333333
 ..111111 ..112222 ..222233 ..333333

*/

char *fr_base64_encode(unsigned char *binary, size_t len)
{
	size_t i=0;
	size_t j=0;
	size_t tmp;
	char *encoded;
	if (len < 1) return NULL;

	tmp = 1 + (len/3)*4;
	if (len%3) tmp+=4;
	encoded = malloc(tmp);
	if (NULL == encoded) return NULL;

	while(i<len) {
		encoded[j] = fr_base64_enc[binary[i] >> 2]; // set 6 bits from 1st
		j++;
		tmp = binary[i] << 4 & 0x30 ; // set 2 bits from 1st
		encoded[j] = fr_base64_enc[ tmp ];
		i++;
		if(i >= len) {
			j++;
			encoded[j++] = '=';
			encoded[j++] = '=';
			break;
		}
		tmp |= binary[i] >> 4 ; // add 4 bits from 2nd
		encoded[j] = fr_base64_enc[ tmp ];
		j++; 
		tmp = binary[i] << 2 & 0x3C; // set remaining 4 bits from 2nd
		encoded[j] = fr_base64_enc[ tmp ];
		i++;
		if(i >= len) {
			j++;
			encoded[j++] = '=';
			break;
		}
		tmp |= binary[i] >> 6; // add 2 bits from 3rd
		encoded[j] = fr_base64_enc[ tmp ];
		j++; 
		encoded[j] = fr_base64_enc[ binary[i] & 0x3F ]; // 6 bits from 3rd
		j++; i++;
	}
	encoded[j] = '\0';
	return encoded;
}
/* ==========================================================
   create base64 decode table
   This function must be called before any call to base64_decode.
*/

void fr_base64_doinit()
{
	int i;

	if (fr_base64_init) return;

	memset(fr_base64_dec, 0xFF, sizeof fr_base64_dec); //0xFF is an invalid value.
	for (i=0;i<sizeof fr_base64_enc; i++) {
		fr_base64_dec[fr_base64_enc[i]] = i;
	}
	fr_base64_init = 1;
	return;
}

/* ==========================================================
   Input: encoded -> Base64 encoded printable string.
   Output: Malloced binary string (can contain \0x00)
           len -> number of bytes in returned string
           Returned binary data must be free()d after use.
   !!! base64_init() must be called before this function is used !!!
   ----------------------------------------------------------
   reverse from base64_encode
   4 bytes -> 3 bytes
   ..111111 ..112222 ..222233 ..333333
   11111111 22222222 33333333
*/
char * fr_base64_decode(
        const char const *encoded,
	size_t *len )
{
	char *decoded;
	size_t encoded_size;
	size_t i;
	size_t j;
	unsigned char tmp1, tmp2;

#ifdef HAVE_PTHREAD_H
	pthread_mutex_lock( &fr_base64_mutx );
#endif /* HAVE_PTHREAD_H */

	if (!fr_base64_init) {
		fr_base64_doinit();
	}

#ifdef HAVE_PTHREAD_H
	pthread_mutex_unlock( &fr_base64_mutx );
#endif /* HAVE_PTHREAD_H */

	encoded_size = strlen((char *)encoded);
	if(encoded_size%4) {
		DEBUG("ERROR: Input buffer size not a multiple of 4\n");
		return NULL; //encoded size must be multiple of 4
	}
	*len = (encoded_size/4 + 1)*3;
	while('=' == encoded[encoded_size-1] && encoded_size > 1) encoded_size--; // skip '=' stuffing

	decoded = malloc(*len);
	if (!decoded) return NULL;

	// for every encoded byte, take lower 6 bits and put them together
	for(i=0,j=0; i < encoded_size - 1; i++) {
		tmp1 = fr_base64_dec[ (unsigned char)encoded[i]   ];
		tmp2 = fr_base64_dec[ (unsigned char)encoded[i+1] ];
		switch (i%4) {
			case 0:
				decoded[j++] = (tmp1 << 2) | ((tmp2 >> 4) & 0x03); // 6 + 2 bits
				break;
			case 1:
				decoded[j++] = (tmp1 << 4) | ((tmp2 >> 2) & 0x0F); // 4 + 4 bits
				break;
			case 2:
				decoded[j++] = (tmp1 << 6) | (tmp2 & 0x3F); // 2 + 6 bits
				break;
		}
	}
	decoded[j] = '\0';
	*len = j;
	
	return decoded;
}
//==========================================================
